/****************************************************************************
 *    This file is part of the program ReformAlign (Reformed Alignments)    *
 *               a profile-based meta-alignment approach                    *
 *                                                                          *
 *    Copyright (C) 2014  Dimitrios Lyras                                   *
 *    e-mail: dimLyras@bio.lmu.de                                           *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 3 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *   along with this program.  If not, see http://www.gnu.org/licenses/.    *
 ****************************************************************************/


#include "FNW.h"

void FNW::GetMotif(Profile& V_Seq, Profile& H_Seq, int SupporterID) {

    //1. Calculate the Distances Matrix //{
    DoubleMat MMat(V_Seq.GetPatternsSize()+1, DoubleVec(H_Seq.GetPatternsSize()+1, 0.0f));
    DoubleMat XMat(V_Seq.GetPatternsSize()+1, DoubleVec(H_Seq.GetPatternsSize()+1, 0.0f));
    DoubleMat YMat(V_Seq.GetPatternsSize()+1, DoubleVec(H_Seq.GetPatternsSize()+1, 0.0f));
    DirMat    MtrMat(V_Seq.GetPatternsSize()+1, DirVec(H_Seq.GetPatternsSize()+1, 0));
    DirMat    XtrMat(V_Seq.GetPatternsSize()+1, DirVec(H_Seq.GetPatternsSize()+1, 0));
    DirMat    YtrMat(V_Seq.GetPatternsSize()+1, DirVec(H_Seq.GetPatternsSize()+1, 0));

    size_t dir;

    dir = GetDistanceMatrix(V_Seq, H_Seq, MMat, XMat, YMat, MtrMat, XtrMat, YtrMat, SubMatMotif);
    //}

    //2. (Optional for Debugging) Print Matrices //{
    //cout<<GetAllDoubleMatsToStr(XMat, YMat, MMat)<<endl;
    //cout<<GetAllDirMatsToStr(XtrMat, YtrMat, MtrMat)<<endl;
    //}

    //3. Traverse Backwards //{
    int i = V_Seq.GetPatternsSize();
    int j = H_Seq.GetPatternsSize();

    map<int, Profile::Pattern> Insertions;

    while((i!=0)||(j!=0)) {
        if (i==0 && dir!=LEFTWARDS) {
            dir = LEFTWARDS;
        }
        if (j==0 && dir!=UPWARDS)   {
            dir = UPWARDS;
        }

        switch(dir) {
        case DIAGONALLY: { //GO DIAGONALLY
            for ( auto H_it = H_Seq.GetPatternAt(j-1).begin(); H_it != H_Seq.GetPatternAt(j-1).end(); ++H_it ) {
                if ( V_Seq.PatternContainsSite(i-1, (*H_it).first ) ) {
                    if ( !V_Seq.IsSiteSupportedBy(i-1, (*H_it).first , SupporterID) ) {
                        V_Seq.IncreaseWeight(i-1, (*H_it).first, 1, SupporterID);
                    }
                }else {
                    V_Seq.InsertSiteAt(i-1, (*H_it).first, 1, SupporterID);
                }
            }

            dir = MtrMat[i][j];
            i--;
            j--;
            break;
        }
        case UPWARDS: { //GO UP
            //Leave profile as it is
            dir = XtrMat[i][j];
            i--;
            break;
        }
        case LEFTWARDS: { //GO LEFT
            Profile::Pattern H_Pat;
            Profile::CreatePatternCopy(H_Seq.GetPatternAt(j-1), H_Pat);
            Insertions[i] = H_Pat;

            //Inserted Positions are not inserted into the V_seq Profile yet, because this would change the V_Seq size
            //leading to erroneous indices during the backwards traversal step
            dir = YtrMat[i][j];
            j--;
            break;
        }
        default:
            cout<<"Unknown Direction!"<<endl;
            exit(1);
        }
    }

    //Now Insert all New Positions to the V_Seq Profile
    for (auto rit=Insertions.rbegin(); rit!=Insertions.rend(); ++rit) {
        V_Seq.InsertPatternAt((*rit).first, (*rit).second);
    }
    //}

}

bool FNW::GetAlignment(Profile& V_Seq, Profile& H_Seq, string& SeqAlignment, string& Alphabet) {

    //1. Calculate the Distances Matrix //{
    DoubleMat MMat(V_Seq.GetPatternsSize()+1, DoubleVec(H_Seq.GetPatternsSize()+1, 0.0f));
    DoubleMat XMat(V_Seq.GetPatternsSize()+1, DoubleVec(H_Seq.GetPatternsSize()+1, 0.0f));
    DoubleMat YMat(V_Seq.GetPatternsSize()+1, DoubleVec(H_Seq.GetPatternsSize()+1, 0.0f));
    DirMat    MtrMat(V_Seq.GetPatternsSize()+1, DirVec(H_Seq.GetPatternsSize()+1, 0));
    DirMat    XtrMat(V_Seq.GetPatternsSize()+1, DirVec(H_Seq.GetPatternsSize()+1, 0));
    DirMat    YtrMat(V_Seq.GetPatternsSize()+1, DirVec(H_Seq.GetPatternsSize()+1, 0));

    size_t dir;

    dir = GetDistanceMatrix(V_Seq, H_Seq, MMat, XMat, YMat, MtrMat, XtrMat, YtrMat, SubMatMotif);
    //}

    //2. (Optional for Debugging) Print Matrices //{
    //cout<<GetAllDoubleMatsToStr(XMat, YMat, MMat)<<endl;
    //cout<<GetAllDirMatsToStr(XtrMat, YtrMat, MtrMat)<<endl;
    //}

    //3. Traverse Backwards //{
    int i = V_Seq.GetPatternsSize();
    int j = H_Seq.GetPatternsSize();
    stringstream ss;
    ss.str();

    while((i!=0)||(j!=0)) {
        if (i==0 && dir!=LEFTWARDS) {
            dir = LEFTWARDS;
        }
        if (j==0 && dir!=UPWARDS)   {
            dir = UPWARDS;
        }

        switch(dir) {
        case DIAGONALLY: { //GO DIAGONALLY
            auto H_it = H_Seq.GetPatternAt(j-1).begin();
            ss<<Alphabet.at((*H_it).first);

            dir = MtrMat[i][j];
            i--;
            j--;
            break;
        }
        case UPWARDS: { //GO UP
            ss<<"-";

            dir = XtrMat[i][j];
            i--;
            break;
        }
        case LEFTWARDS: { //GO LEFT
            return true;
            //Automatically Break the loop and return that Fine-Tuning is required as soon as an Insertion
            //is detected to be required in the Profile
            dir = YtrMat[i][j];
            j--;
            break;
        }
        default:
            cout<<"Unknown Direction!"<<endl;
            exit(1);
        }
    }
    //}

    //4. Assign the "Reversed" Alignment to the given argument and return false (no Fine tuning is required) //{
    string TempAlignStr = ss.str();
    std::reverse(TempAlignStr.begin(), TempAlignStr.end());
    SeqAlignment = std::move(TempAlignStr);
    return false;
    //}

}

double FNW::GetDistanceMatrix(Profile& V_Seq, Profile& H_Seq, DoubleMatRef MMat, DoubleMatRef XMat, DoubleMatRef YMat, DirMatRef MtrMat, DirMatRef XtrMat, DirMatRef YtrMat, DoubleMatRef SubMat) {
    size_t rows = V_Seq.GetPatternsSize()+1;
    size_t cols = H_Seq.GetPatternsSize()+1;

    //1. Initialize the 1st Row and 1st Column of all Tables //{
    XMat[0][1] = XMat[1][0] = YMat[0][1] = YMat[1][0] = MMat[0][1] = MMat[1][0] = TermGapOpPenalty;
    MtrMat[0][1] = XtrMat[0][1] = YtrMat[0][1] = LEFTWARDS;
    MtrMat[1][0] = XtrMat[1][0] = YtrMat[1][0] = UPWARDS;

    for (unsigned int i=2; i<rows; i++) {
        MMat[i][0] = MMat[i-1][0] + TermGapExPenalty;
        XMat[i][0] = XMat[i-1][0] + TermGapExPenalty;
        YMat[i][0] = YMat[i-1][0] + TermGapExPenalty;

        MtrMat[i][0] = UPWARDS;
        XtrMat[i][0] = UPWARDS;
        YtrMat[i][0] = UPWARDS;
    }

    for (unsigned int j=2; j<cols; j++) {
        MMat[0][j] = MMat[0][j-1] + TermGapExPenalty;
        XMat[0][j] = XMat[0][j-1] + TermGapExPenalty;
        YMat[0][j] = YMat[0][j-1] + TermGapExPenalty;

        MtrMat[0][j] = LEFTWARDS;
        XtrMat[0][j] = LEFTWARDS;
        YtrMat[0][j] = LEFTWARDS;
    }
    //}

    //2. Populate all the remaining cells of the Tables //{
    for (unsigned int i=1; i<rows; i++ ) {
        for (unsigned int j=1; j<cols; j++ ) {
            //Substitution Score Calculation - The Strict Scheme //{
            double TotSubScore = 0.0f;

            Profile::PatternRef H_Pat = H_Seq.GetPatternAt(j-1);
            Profile::PatternRef V_Pat = V_Seq.GetPatternAt(i-1);

            for ( auto H_it = H_Pat.begin(); H_it != H_Pat.end(); ++H_it ) {
                for ( auto V_it = V_Pat.begin(); V_it != V_Pat.end(); ++V_it ) {
                    TotSubScore += ((*H_it).second.Weight * (*V_it).second.Weight) * SubMat[(*H_it).first][(*V_it).first];
                }
            }
            //}

            double val1, val2, val3;
            //double Pattern_Uniformity = V_Seq.GetIdentityAt(i-1, (*H_Pat.begin()).first );
            double Pattern_Uniformity = V_Seq.GetMaxIdentityAt(i-1);
            val1 = MMat[i-1][j] + VGapOpPenalty - (TotSubScore*Pattern_Uniformity);
            val2 = XMat[i-1][j] + VGapExPenalty;
            if (val1>=val2) {
                XMat[i][j]   = val1;
                XtrMat[i][j] = DIAGONALLY;
            } else {
                XMat[i][j]   = val2;
                XtrMat[i][j] = UPWARDS; //from the XMat
            }

            val1 = MMat[i][j-1] + HGapOpPenalty - (TotSubScore*Pattern_Uniformity);
            val2 = YMat[i][j-1] + HGapExPenalty;
            if (val1>=val2) {
                YMat[i][j]   = val1;
                YtrMat[i][j] = DIAGONALLY;
            } else {
                YMat[i][j]   = val2;
                YtrMat[i][j] = LEFTWARDS; //from the YMat
            }

            val1 = MMat[i-1][j-1] + TotSubScore;
            val2 = XMat[i-1][j-1] + TotSubScore;
            val3 = YMat[i-1][j-1] + TotSubScore;
            if (val1>=val2 && val1>=val3) {
                MMat[i][j]   = val1;
                MtrMat[i][j] = DIAGONALLY;
            } else if (val2>=val1 && val2>=val3) {
                MMat[i][j]   = val2;
                MtrMat[i][j] = UPWARDS; //from the XMat
            } else {
                MMat[i][j]   = val3;
                MtrMat[i][j] = LEFTWARDS; //from the YMat
            }
        }
    }
    //}

    //3. After All Cells have been computed, decide and return the lower-right cell starting direction (for backwards traversal) //{
    if (MMat[rows-1][cols-1]>=XMat[rows-1][cols-1] && MMat[rows-1][cols-1]>=YMat[rows-1][cols-1]) {
        return DIAGONALLY;
    } else if (XMat[rows-1][cols-1]>=MMat[rows-1][cols-1] && XMat[rows-1][cols-1]>=YMat[rows-1][cols-1]) {
        return UPWARDS;
    } else {
        return LEFTWARDS;
    }
    //}

}


string FNW::GetDoubleMatToStr(DoubleMatRef Mat) {
    stringstream ss;
    ss.str("");

    ss<<"\n\tVALUES MATRIX\n\t=============\n";
    ss<<endl;
    ss.setf(std::ios::fixed);
    ss.precision(2);
    for (unsigned int i=0; i<Mat.size(); i++) {
        for (unsigned int j=0; j<Mat[i].size(); j++) {
            ss<<left<<setw(PRINTWIDTH)<<Mat[i][j];
        }
        ss<<endl;
    }

    return ss.str();
}

string FNW::GetDirMatToStr(DirMatRef Mat) {
    stringstream ss;
    ss.str("");

    ss<<"\n\tDIRECTIONS MATRIX\n\t=================\n";
    ss<<endl;
    ss.setf(std::ios::fixed);
    ss.precision(2);
    for (unsigned int i=0; i<Mat.size(); i++) {
        for (unsigned int j=0; j<Mat[i].size(); j++) {
            ss<<left<<setw(PRINTWIDTH)<<Mat[i][j];
        }
        ss<<endl;
    }

    return ss.str();
}

string FNW::GetAllDoubleMatsToStr(DoubleMatRef XMat, DoubleMatRef YMat, DoubleMatRef MMat) {
    stringstream ss;
    ss.str("");

    ss<<"\n\tALL VALUE MATRICES\n\t==================\n";
    ss<<endl;
    ss.setf(std::ios::fixed);
    ss.precision(2);

    for (unsigned int i=0; i<MMat.size(); i++) {
        for (unsigned int j=0; j<MMat[i].size(); j++) {
            ss<<left<<setw(PRINTWIDTH)<<MMat[i][j]<<left<<setw(PRINTWIDTH+2)<<XMat[i][j];
        }
        ss<<endl;
        for (unsigned int j=0; j<MMat[i].size(); j++) {
            ss<<left<<setw(PRINTWIDTH)<<YMat[i][j]<<setw(PRINTWIDTH+2)<<"";
        }
        ss<<endl<<endl;
    }
    return ss.str();
}

string FNW::GetAllDirMatsToStr(DirMatRef XtrMat, DirMatRef YtrMat, DirMatRef MtrMat) {
    stringstream ss;
    ss.str("");

    ss<<"\n\tALL DIRECTION MATRICES\n\t======================\n";
    ss<<endl;
    ss.setf(std::ios::fixed);
    ss.precision(2);

    for (unsigned int i=0; i<MtrMat.size(); i++) {
        for (unsigned int j=0; j<MtrMat[i].size(); j++) {
            ss<<left<<setw(PRINTWIDTH)<<MtrMat[i][j]<<left<<setw(PRINTWIDTH+2)<<XtrMat[i][j];
        }
        ss<<endl;
        for (unsigned int j=0; j<MtrMat[i].size(); j++) {
            ss<<left<<setw(PRINTWIDTH)<<YtrMat[i][j]<<setw(PRINTWIDTH+2)<<"";
        }
        ss<<endl<<endl;
    }

    return ss.str();
}

string FNW::GetParamsToStr(string& Alphabet) {
    stringstream ss; ss.str("");
    ss << endl;
    ss << "\t\tPARAMETERS\n\t\t==========\n\n";
    ss << "Residue Weight Matrix (Motifs):\n";

    for (unsigned int i=0; i<SubMatMotif.size(); i++) {
        if (i==0) {
            ss << setw(4) << " ";
        }

        ss << setw(4) <<Alphabet.at(i);
    }
    ss<<endl;

    for (unsigned int i=0; i<SubMatMotif.size(); i++) {
        for (unsigned int j=0; j<SubMatMotif[i].size(); j++) {
            if (j==0) {
                ss<<setw(4)<<Alphabet.at(i);
            }
            ss<<setw(4)<<SubMatMotif[i][j];
        }
        ss<<endl;
    }


    ss << "\nResidue Weight Matrix (Aligning):\n";

    for (unsigned int i=0; i<SubMatAlign.size(); i++) {
        if (i==0) {
            ss << setw(4) << " ";
        }
        ss << setw(4) <<Alphabet.at(i);
    }
    ss<<endl;

    for (unsigned int i=0; i<SubMatAlign.size(); i++) {
        for (unsigned int j=0; j<SubMatAlign[i].size(); j++) {
            if (j==0) {
                ss<<setw(4)<<Alphabet.at(i);
            }
            ss<<setw(4)<<SubMatAlign[i][j];
        }
        ss<<endl;
    }

    ss<<endl<<"  H. Gap Opening Penalty : "<< HGapOpPenalty<<endl;
    ss<<" H.Gap Extension Penalty : "<< HGapExPenalty<<endl;
    ss<<" V.Gap Opening   Penalty : "<< VGapOpPenalty<<endl;
    ss<<" V.Gap Extension Penalty : "<< VGapExPenalty<<endl;
    ss<<"Terminal Gap Op. Penalty : "<< TermGapOpPenalty<<endl;
    ss<<"Terminal Gap Ex. Penalty : "<< TermGapExPenalty<<endl;

    return ss.str();
}

